# PipeLink

This library is intended to simplify the process of getting structured data
from one process to another, using pipes, sockets, or a custom transport.


## Installing

Install using composer. Until PipeLink is added to Packagist you will have to
follow the appropriate instructions to add the repository.

```
composer require noccylabs/pipe-link:^0.1.0
```

## Scope

PipeLink provides the low-level interface, i.e. basic session control and
message marshalling. Everything above this should be implemented in a wrapper
with a higher level API.

```php

// This is a demonstration, not an example! Write an appropriate
// high-level API around PipeLink for your code to use.
class ThingWrapper {

  private LinkInterface $link;

  //...

  public function ping() {
    $this->link->sendMessage("ping", null);
    $this->pongTimeout = microtime(true) + 3;
  }

  //...

  public function update() {
    while ($msg = $this->readMessage()) { // note: non-blocking
      if ($msg->getType() == "pong") {
       // ...
       $this->pongTimeout = null;
      }
    }
    if ($this->pongTimeout !== null && microtime(true) > $this->pongTimeout) {
      // timeout!
    }
  }

}
```

## Usage

There are two general sitatuions you will encounter; having a pair of streams such as
standard input and standard output, or having a single stream for both reading and
writing, such as a socket.

```php
// Use the createFromPipes factory method if you have a pair:
$link = Link::createFromPipes(STDIN, STDOUT);
// Or if you want to save a few bytes:
$link = Link::createFromStandardIo();

// On the other hand, if you have a single stream:
$link = Link::createFromStream($socket);

// Or if you intend to pcntl_fork() you can prepare a pair:
[$a,$b] = Link::createPair();
$pid = pcntl_fork();
if ($pid === 0) {
  // use the $b link from child
} elseif ($pid > 0) {
  // use the $a link from parent
}
```

Now that you have a link, you can either use it as it is, i.e. send and receive messages
as you see fit without any added unicorn dust, or you can add some magic to maket things
easier for yourself.

### Basic use

For a basic example, take a look at this:

```php
// Blocking example, note the parameter to readMessage
while ($msg = $link->readMessage(true)) {
  switch ($msg->getType()) {
    case "echo":
      $response = new Message(
        type:"echo.response", 
        body:$msg->getBody(), 
        responseToId:$msg->getId()
      );
      $link->sendMessage($response);
      break;
    // ..
  }
}

// Asynchronous example, if you have things to do
while ($running){
  // You don't have to call hasMessage() with non-blocking readMessage()
  // but including it for clarity.
  if ($link->hasMessage()) {
    $msg = $link->readMessage();
    handle_msg($msg);
  }
  // Do your stuff here
  do_stuff();
  // Don't forget to chill
  usleep(10000);
}
```

### Hooking messages

You can also hook messages using `onMessage`

```php
// Hook some interesting messages
$link->onMessage("echo", function (Message $message, LinkInterface $link) {
  $response = new Message(
    type:"echo.response",
    body:$message->getBody(),
    responseToId:$message->getId()
  );
  // Returning a response will send it for us
  return $response;
}, false);

while ($msg = $link->readMessage(true)) {
  // the "echo" messages will never show up here, as the relay parameter (#3)
  // to onMessage() is false. We still need to call readMessage() or at least
  // hasMessage() to keep reading and notifying tho.
}
$link->hasMessage();
```

### Adding processing

```php
// You can easily filter and process incoming messages by 
$link->addInputFilter(function (Message $message) {
  // This is invoked for every message received
  if ($message->getType() == "evilmessage") {
    // Return false to drop this message
    return false;
  }
  if ($message->getType() == "usermessage") {
    // Return a message or derived class to replace the message
    return new UserMessage($message);
  }
});

// You can do the same for outgoing messages:
$link->addOutputFilter(function (Message $message) {
  $message->setAttribute("processId", "worker42");
  $message->setAttribute("authorization", "MagicCookie");
});

// Sending a messge will add the attributes:
$link->sendMessage(new Message(type:"msgtype", body:"data"));
// Unless the message is sent "direct", by setting 2nd parameter to true
$link->sendMessage(new Message(type:"msgtype", body:"not filtered"), true);

```

### Using the mapper

```php

class PingMessage extends Message {}

class PongMessage extends Message {}

$link->addInputFilter(new MessageMapper([
  'ping' => PingMessage::class,
  'pong' => PongMessage::class,
]));

while ($msg = $link->readMessage(true)) {
  if ($msg instanceof PingMessage) {
    // ...
  } elseif ($msg instanceof PongMessage) {
    // ...
  } else {
    // ...
  }
}

```
