<?php

namespace NoccyLabs\PipeLink;
use ArrayAccess;
use JsonSerializable;
use SplQueue;

/**
 *
 *
 *
 */
abstract class Link
{

    private array $inputHandlers = [];

    private array $outputHandlers = [];

    private array $messageHandlers = [];

    protected $inputQueue = [];

    /**
     * @return LinkInterface
     */
    public static function createFromStandardIo(): LinkInterface
    {
        return StreamLink::createFromPipes(STDIN, STDOUT);
    }

    /**
     * @return LinkInterface
     */
    public static function createFromPipes($inPipe, $outPipe): LinkInterface
    {
        return new StreamLink($inPipe, $outPipe);
    }

    /**
     * @return LinkInterface
     */
    public static function createFromStream($ioStream): LinkInterface
    {
        return new StreamLink($ioStream, $ioStream);
    }


    /**
     * @return array<int,LinkInterface>
     */
    public static function createPair(): array
    {
        if (!$pair = stream_socket_pair(STREAM_PF_UNIX, STREAM_SOCK_STREAM, STREAM_IPPROTO_IP)) {
            throw new \RuntimeException("Unable to create socket pair");
        }
        return [
            new StreamLink($pair[0], $pair[0]),
            new StreamLink($pair[1], $pair[1])
        ];
    }


    public function sendMessage(Message $message, bool $direct=false): void
    {
        // Pass the message through the handlers, unless $direct is set
        if ($direct === false) {
            $message = $this->processOutgoingMessage($message);
            if ($message === false) return;
        }

        // Pack the message and send it to output
        $packedMessage = $message->pack();
        $this->writeFrame($packedMessage);
    }

    public function readMessage(bool $block=false): ?Message
    {
        if ($block) {
            while (!$this->hasMessage()) {
                usleep(10000);
            }
            return array_shift($this->inputQueue);
        } else {
            if ($this->hasMessage()) {
                return array_shift($this->inputQueue);
            }
            return null;
        }
    }

    public function hasMessage(): bool
    {
        $this->update();
        return (count($this->inputQueue) > 0);
    }

    public function onMessage(string $type, callable $handler, bool $relay=false): void
    {
        if (!array_key_exists($type, $this->messageHandlers)) {
            $this->messageHandlers[$type] = [];
        }
        $this->messageHandlers[$type][] = [ $handler, $relay ];
    }

    public function addInputHandler(callable $handler, bool $prepend=false): void
    {
        if ($prepend) {
            array_unshift($this->inputHandlers, $handler);
        } else {
            array_push($this->inputHandlers, $handler);
        }
    }

    public function addOutputHandler(callable $handler, bool $prepend=false): void
    {
        if ($prepend) {
            array_unshift($this->outputHandlers, $handler);
        } else {
            array_push($this->outputHandlers, $handler);
        }
    }

    abstract public function update(): void;

    abstract protected function writeFrame(string $frame): void;

    protected function pushInputMessage(Message $message): void
    {
        $message = $this->processIncomingMessage($message);
        if ($message === false) return;
        // Call on the message handlers, starting with exact matches
        $messageType = $message->getType();
        if (array_key_exists($messageType, $this->messageHandlers)) {
            foreach ($this->messageHandlers[$messageType] as [$cb,$relay]) {
                $result = call_user_func($cb, $message, $this);
                if ($result instanceof Message) {
                    $this->sendMessage($result);
                }
                if ($relay === false) return;
            }
        }
        // Determine if there are any wildcard matches
        $wildcards = array_filter(
            array_keys($this->messageHandlers),
            fn($h) => fnmatch($h, $messageType) 
        );
        foreach ($wildcards as $key) {
            foreach ($this->messageHandlers[$key] as [$cb,$relay]) {
                $result = call_user_func($cb, $message, $this);
                if ($result instanceof Message) {
                    $this->sendMessage($result);
                }
                if ($relay === false) return;
            }
        }

        // Stash the message
        $this->inputQueue[] = $message;
    }

    private function processOutgoingMessage(Message $message): Message|bool
    {
        foreach ($this->outputHandlers as $handler) {
            $handlerResult = call_user_func($handler, $message);
            if ($handlerResult === false) return false;
            if ($handlerResult === true) return $message;
            if ($handlerResult instanceof Message) $message = $handlerResult;
        }
        return $message;
    }

    private function processIncomingMessage(Message $message): Message|bool
    {
        foreach ($this->inputHandlers as $handler) {
            $handlerResult = call_user_func($handler, $message);
            if ($handlerResult === false) return false;
            if ($handlerResult === true) return $message;
            if ($handlerResult instanceof Message) $message = $handlerResult;
        }
        return $message;
    }

}

