<?php

namespace NoccyLabs\PipeLink;
use ArrayAccess;
use JsonSerializable;
use SplQueue;

class Message implements JsonSerializable
{

    protected string $type;

    protected mixed $body;

    protected array $attributes = [];

    protected ?string $msgId = null;

    protected ?string $responseToId = null;


    public static function unpack(string &$buffer): ?Message
    {
        $metaLen = 2;

        if (strlen($buffer) < $metaLen) {
            return null;
        }

        $meta = unpack("nlength", substr($buffer, 0, $metaLen));

        $frameLen = 2 + $meta['length'];
        if (strlen($buffer) < $frameLen) {
            return null;
        }

        // Extract the frame from the buffer and parse the body
        $frame = substr($buffer, 0, $frameLen);
        $buffer = substr($buffer, $frameLen);
        $body = json_decode(substr($frame, 2, $meta['length']), true);

        // Sanity checks
        if (empty($body['type']) || !array_key_exists('body',$body)) {
            throw new \RuntimeException("Invalid message received");
        }

        // Build the object
        $message = new Message(
            type: $body['type'],
            body: $body['body'],
            attributes: $body['attr']??[],
            msgId: $body['mid']??null,
            responseToId: $body['rto']??null,
        );

        return $message;
    }

    public function __construct(string $type, mixed $body, array $attributes = [], ?string $msgId = null, ?string $responseToId = null)
    {
        $this->type = $type;
        $this->body = $body;
        $this->attributes = $attributes;
        $this->msgId = $msgId;
        $this->responseToId = $responseToId;
    }

    public function getType(): string
    {
        return $this->type;
    }

    public function getBody(): mixed
    {
        return $this->body;
    }

    public function setBody(mixed $body): void
    {
        $this->body = $body;
    }

    public function get(string $path, mixed $default=null): mixed
    {
        $pathSegments = explode(".", ltrim($path, "."));
        $pathPointer = $this->body;

        while ($pathItem = array_shift($pathSegments)) {
            if (!is_array($pathPointer)) return $default;
            if (array_key_exists($pathItem, $pathPointer)) {
                $pathPointer = $pathPointer[$pathItem];
            }
        }
        return $pathPointer;
    }

    public function getMessageId(): ?string
    {
        return $this->msgId;
    }

    public function hasAttribute(string $name): bool
    {
        return array_key_exists($name, $this->attributes);
    }

    public function getAttribute(string $name, mixed $default=null): mixed
    {
        return $this->attributes[$name]??$default;
    }

    public function setAttribute(string $name, mixed $value): void
    {
        if ($value === null) {
            unset($this->attributes[$name]);
            return;
        }
        $this->attributes[$name] = $value;
    }

    public function pack(): string
    {
        $json = json_encode($this);
        $jsonLen = strlen($json);

        $head = pack("n", $jsonLen);

        return $head.$json;
    }

	/**
	 * Specify data which should be serialized to JSON
	 * Serializes the object to a value that can be serialized natively by json_encode().
	 * @return mixed Returns data which can be serialized by json_encode(), which is a value of any type other than a resource .
	 */
	public function jsonSerialize()
    {
        $json = [
            'type' => $this->type,
            'body' => $this->body
        ];
        if (count($this->attributes) > 0) $json['attr'] = $this->attributes;
        if ($this->msgId) $json['mid'] = $this->msgId;
        if ($this->responseToId) $json['rto'] = $this->responseToId;
        return $json;
	}
}

